import re
from pyprove import expres

def solo(pid, name, mult=0, noinit=False, efun="Enigma"):
   proto = expres.protos.load(pid)
   enigma = "1*%s(PreferWatchlist,%s,%s)" % (efun, name, mult)
   eproto = "%s-H'(%s)'" % (proto[:proto.index("-H'")], enigma)
   post = efun
   post += ("0M%s" % mult) if mult else "0"
   if noinit:
      eproto = eproto.replace("--prefer-initial-clauses", "")
      post += "No" 
   epid = "Enigma+%s+%s+%s" % (name.replace("/","+"), pid, post)
   expres.protos.save(epid, eproto)
   return epid

def coop(pid, name, freq=None, mult=0, noinit=False, efun="Enigma"):
   proto = expres.protos.load(pid)
   post = efun
   if not freq:
      freq = sum(map(int,re.findall(r"(\d*)\*", proto)))
      post += "S"
   else:
      post += "F%s"% freq
   post += ("M%s" % mult) if mult else ""
   enigma = "%d*%s(PreferWatchlist,%s,%s)" % (freq,efun,name,mult)
   eproto = proto.replace("-H'(", "-H'(%s,"%enigma)
   if noinit:
      eproto = eproto.replace("--prefer-initial-clauses", "")
      post += "No"
   epid = "Enigma+%s+%s+%s" % (name.replace("/","+"), pid, post)
   expres.protos.save(epid, eproto)
   return epid




import sys
import numpy
import xgboost as xgb
from scipy.sparse import csr_matrix, lil_matrix
from sklearn.model_selection import train_test_split
from sklearn.datasets import load_svmlight_file


def train(f_in, f_out, log=None, xgb_params=None):
   if log:
      log.write("\nTraining XGBoost model (%s):\n\n" % f_in)
      oldout = sys.stdout
      sys.stdout = log

   dtrain = xgb.DMatrix(f_in)

   labels = dtrain.get_label()
   pos = float(len([x for x in labels if x == 1]))
   neg = float(len([x for x in labels if x == 0]))

   param = {
      'max_depth': 9, 
      'eta': 0.3, 
      'objective': 'binary:logistic', 
      'scale_pos_weight': (neg/pos),
   }
   if xgb_params:
      param.update(xgb_params)
   if "num_round" in param:
      num_round = param["num_round"]
      del param["num_round"]
   else:
      num_round = 200

   bst = xgb.train(param, dtrain, num_round, evals=[(dtrain, "training")])
   bst.save_model(f_out)

   if log:
      sys.stdout = oldout

   return bst

def train_old(f_in, f_out, log=None):
   if log:
      log.write("\nTraining XGBoost model (%s):\n\n" % f_in)
      oldout = sys.stdout
      sys.stdout = log
   
   X, y = load_svmlight_file(f_in)
   X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)
   #X_all, X_none, y_all, y_none = train_test_split(X, y, test_size=0.0, random_state=42)

   c_pos = len([y for y in y_test if y == 1])
   c_neg = len([y for y in y_test if y == 0])
   cols = X_train.shape[1]
   X_pos = lil_matrix((c_pos, cols), dtype=numpy.float64)
   X_neg = lil_matrix((c_neg, cols), dtype=numpy.float64)

   i = 0
   j = 0
   for (row, label) in zip(X_test, y_test):
      if label == 1:
         X_pos[i] = row
         i += 1
      elif label == 0:
         X_neg[j] = row
         j += 1

   y_pos = numpy.empty(c_pos)
   y_pos.fill(1)
   y_neg = numpy.empty(c_neg)
   y_neg.fill(0)

   dtest = xgb.DMatrix(X_test, label=y_test)
   dpos = xgb.DMatrix(X_pos, label=y_pos)
   dneg = xgb.DMatrix(X_neg, label=y_neg)

   dtrain = xgb.DMatrix(X_train, label=y_train)
   #dtrain = xgb.DMatrix(X_all, label=y_all)
   param = {'max_depth':9, 'eta':0.3, 'objective':'binary:logistic'}
   num_round = 6
   #num_round = 200

   bst = xgb.train(param, dtrain, num_round, evals=[(dtrain, "training"), (dtest,"testing"), (dpos, "pos"), (dneg, "neg")])
   bst.save_model(f_out)

   if log:
      sys.stdout = oldout

   return bst




import lightgbm as lgb
from .learner import Learner

DEFAULTS = {
   'max_depth': 9, 
   'learning_rate': 0.3, 
   'objective': 'binary', 
   'num_round': 200,
   'num_leaves': 300
}

class LightGBM(Learner):

   def __init__(self, **args):
      self.params = dict(DEFAULTS)
      self.params.update(args)

   def efun(self):
      return "EnigmaLgb"

   def ext(self):
      return "lgb"

   def name(self):
      return "LightGBM"

   def __repr__(self):
      args = ["%s=%s"%(x,self.params[x]) for x in self.params]
      args = ", ".join(args)
      return "%s(%s)" % (self.name(), args)

   def train(self, f_in, f_mod):
      dtrain = lgb.Dataset(f_in)
      dtrain.construct()
      labels = dtrain.get_label()
      pos = float(len([x for x in labels if x == 1]))
      neg = float(len([x for x in labels if x == 0]))

      self.params["scale_pos_weight"] = (neg/pos)
      bst = lgb.train(self.params, dtrain, valid_sets=[dtrain])
      bst.save_model(f_mod)
      return bst




import os

def sdbm(s):
   "Generic purpose string hash function (see http://www.cse.yorku.ca/~oz/hash.html)"
   h = 0
   for c in s:
      h = ord(c) + (h << 6) + (h << 16) - h
      h &= 0xFFFFFFFFFFFFFFFF
   return h

def fhash(s, base, cache={}):
   if s not in cache:
      cache[s] = 1 + (sdbm(s) % base)
   return cache[s]
      
def load(f_map):
   emap = {}
   if not os.path.exists(f_map):
      return emap
   for line in file(f_map):
      if line.startswith("version") or line.startswith("hash_base"):
         continue
      # FIXME: load hashed stuff
      (fid,ftr) = line.strip().split("(")[1].split(",")
      fid = int(fid.strip(", "))
      ftr = ftr.strip('") .')
      emap[ftr] = fid
   return emap

def save(emap, f_map, version, hashing=None):
   out = file(f_map, "w")
   out.write('version("%s").\n' % version)
   if hashing:
      out.write('hash_base(%s).\n' % hashing)
      # informative hash dump to see collisions
      for x in sorted(emap):
         out.write('bucket(%s, "%s").\n' % (emap[x], x))
   else:
      rev = {emap[ftr]:ftr for ftr in emap}
      for x in sorted(rev):
         out.write('feature(%s, "%s").\n' % (x,rev[x]))
   out.close()

def create(pre, hashing=None):

   def add(features, new):
      for feature in new.strip().split(" "):
         if (not feature) or feature.startswith("$"):
            continue
         if "/" in feature:
            feature = feature.split("/")[0]
         features.add(feature)

   features = set()
   for pr in pre:
      (sign,clause,conj) = pr.strip().split("|")
      add(features, clause)
      add(features, conj)

   if hashing:
      return {f:fhash(f,hashing) for f in sorted(features)}
   else:
      return {y:x for (x,y) in enumerate(sorted(features), start=1)}

def join(f_maps):
   features = set()
   for f_map in f_maps:
      features.update(load(f_map).keys())
   return {y:x for (x,y) in enumerate(sorted(features), start=1)}




import os
from . import enigmap, pretrains, trains, protos
from pyprove import expres, log

ENIGMA_ROOT = os.getenv("ENIGMA_ROOT", "./Enigma")

DEFAULTS = {
   "gzip": True,
   "force": False,
   "hashing": None,
   "version": "VHSLC",
   "eargs": "--training-examples=3 -s",
   "cores": 4,
   "hash_debug": False,
}


def path(name, filename=None):
   if filename:
      return os.path.join(ENIGMA_ROOT, name, filename)
   else:
      return os.path.join(ENIGMA_ROOT, name)


def collect(name, rkeys, settings):
   version = settings["version"]
   force = settings["force"]
   cores = settings["cores"]
   hashing = settings["hashing"] if not settings["hash_debug"] else None

   f_dat = path(name, "train.%s" % ("in" if hashing else "pre"))
   if force or not os.path.isfile(f_dat):
      log.msg("+ extracting training data from results")
      pretrains.prepare(rkeys, version, force, cores, hashing)
      log.msg("+ collecting %s data" % ("training" if hashing else "pretrains"))
      pretrains.make(rkeys, out=file(f_dat, "w"), hashing=hashing)


def setup(name, rkeys, settings):
   os.system("mkdir -p %s" % path(name))
   f_pre = path(name, "train.pre")
   f_map = path(name, "enigma.map")
   f_log = path(name, "train.log")
   hashing = settings["hashing"]
  
   if os.path.isfile(f_map) and os.path.isfile(f_pre) and not settings["force"]:
      return enigmap.load(f_map) if not hashing else hashing
      
   if rkeys:
      collect(name, rkeys, settings)

   if hashing and not settings["hash_debug"]:
      file(f_map,"w").write('version("%s").\nhash_base(%s).\n' % (settings["version"], hashing))
      return hashing

   #if os.path.isfile(f_log):
   #   os.system("rm -f %s" % f_log)
   if settings["force"] or not os.path.isfile(f_map):
      log.msg("+ creating feature info")
      emap = enigmap.create(file(f_pre), hashing)
      enigmap.save(emap, f_map, settings["version"], hashing)
   else:
      if not hashing:
         emap = enigmap.load(f_map)

   return emap if not hashing else hashing


def make(name, rkeys, settings):
   learner = settings["learner"]

   f_pre = path(name, "train.pre")
   f_in  = path(name, "train.in")
   f_mod = path(name, "model.%s" % learner.ext())
   f_log = path(name, "train.log")

   if os.path.isfile(f_mod) and not settings["force"]:
      return True

   emap = setup(name, rkeys, settings)
   if not emap:
      os.system("rm -fr %s" % path(name))
      return False
   
   if settings["hash_debug"] or not settings["hashing"]:
      if settings["force"] or not os.path.isfile(f_in):
         log.msg("+ generating training data")
         trains.make(file(f_pre), emap, out=file(f_in, "w"))

   log.msg("+ training %s" % learner.name())
   tlog = file(f_log, "a")
   learner.build(f_in, f_mod, tlog)
   tlog.close()

   if settings["gzip"]:
      log.msg("+ compressing training files")
      os.system("cd %s; gzip -qf *.pre *.in *.out 2>/dev/null" % path(name))

   return True


def check(settings):
   if ("h" in settings["version"] and not settings["hashing"]) or \
      (settings["hashing"] and "h" not in settings["version"]):
         raise Exception("enigma.models: Parameter hashing must be set to the hash base (int) iff version contains 'h'.")   
   for x in DEFAULTS:
      if x not in settings:
         settings[x] = DEFAULTS[x]
   for x in ["bid", "pids", "learner"]:
      if x not in settings:
         raise Exception("enigma.models: Required setting '%s' not set!" % x)   
   if "results" not in settings:
      settings["results"] = {}


def update(settings, pids=None):
   if not pids:
      pids = settings["pids"]
   settings["results"].update(expres.benchmarks.eval(
      settings["bid"], pids, settings["limit"], cores=settings["cores"], 
      eargs=settings["eargs"], force=settings["force"]))


def loop(model, settings, nick=None):
   check(settings)
   if nick:
      model = "%s/%s" % (model, nick)
   log.msg("Building model %s" % model)

   update(settings)
   if not make(model, settings["results"], settings):
      raise Exception("Enigma: FAILED: Building model %s" % model)
   efun = settings["learner"].efun()
   new = [
      protos.solo(settings["pids"][0], model, mult=0, noinit=True, efun=efun),
      protos.coop(settings["pids"][0], model, mult=0, noinit=True, efun=efun)
   ]
   settings["pids"].extend(new)
   update(settings, new)
   
   log.msg("Building model finished\n")
   return new





import subprocess
from . import enigmap

#PERF = ["perf", "stat", "-e", "task-clock:up,page-faults:up,instructions:up"]
PERF = []

def train(f_in, f_mod, f_out=None, f_log=None):
   if f_log:
      subprocess.call(PERF+["train", "-s" , "2", f_in, f_mod], \
         stdout=file(f_log, "a"), stderr=subprocess.STDOUT)
   else:
      subprocess.call(["train", "-s" , "2", f_in, f_mod])

   if f_out:
      predict(f_in, f_mod, f_out, f_log)

def predict(f_in, f_mod, f_out, f_log):
   if f_log:
      subprocess.call(PERF+["predict", f_in, f_mod, f_out], \
         stdout=file(f_log, "a"), stderr=subprocess.STDOUT)
   else:
      subprocess.call(["predict", f_in, f_mod, f_out])

def stats(f_in, f_out):
   ins = file(f_in).read().strip().split("\n")
   ins = [int(x.split()[0]) for x in ins]

   outs = file(f_out).read().strip().split("\n")
   outs = map(int, outs)

   stat = {}
   stat["COUNT"] = 0
   stat["COUNT:POS"] = 0
   stat["COUNT:NEG"] = 0
   stat["RIGHT"] = 0
   stat["RIGHT:POS"] = 0
   stat["RIGHT:NEG"] = 0
   stat["WRONG"] = 0
   stat["WRONG:POS"] = 0
   stat["WRONG:NEG"] = 0

   for (correct,predicted) in zip(ins, outs):
      stat["COUNT"] += 1
      if correct == 1:
         stat["COUNT:POS"] += 1
      else:
         stat["COUNT:NEG"] += 1

      if correct == predicted:
         stat["RIGHT"] += 1
         if correct == 1:
            stat["RIGHT:POS"] += 1
         else:
            stat["RIGHT:NEG"] += 1
      else:
         stat["WRONG"] += 1
         if correct == 1:
            stat["WRONG:POS"] += 1
         else:
            stat["WRONG:NEG"] += 1
   
   stat["ACC"] = stat["RIGHT"] / float(stat["COUNT"])
   stat["ACC:POS"] = (stat["RIGHT:POS"] / float(stat["COUNT:POS"])) if int(stat["COUNT:POS"]) != 0 else 0
   stat["ACC:NEG"] = (stat["RIGHT:NEG"] / float(stat["COUNT:NEG"])) if int(stat["COUNT:NEG"]) != 0 else 0

   return stat

def load(f_mod, f_map):
   emap = enigmap.load(f_map)
   lines = file(f_mod)
   rev = {emap[ftr]:ftr for ftr in emap}

   header = {}
   for line in lines:
      line = line.strip()
      if line == "w":
         break
      parts = line.split()
      header[parts[0]] = " ".join(parts[1:])

   # clause features
   w1 = {}
   i = 1
   for line in lines:
      w = float(line.strip())
      w1[rev[i]] = w
      if i == len(emap):
         break
      i += 1

   # conjecture features
   w2 = {}
   i = 1
   for line in lines:
      w = float(line.strip())
      w2[rev[i]] = w
      if i == len(emap):
         break
      i += 1

   lines.close()
   return (header,w1,w2)

def save(header, w1, w2, emap, f_mod):
   rev = {emap[ftr]:ftr for ftr in emap}
   lines = file(f_mod, "w")

   info = lambda key: lines.write("%s %s\n" % (key, header[key]))
   info("solver_type")
   info("nr_class")
   info("label")
   lines.write("nr_feature %d\n" % (2*len(emap)))
   info("bias")
   lines.write("w\n")

   for i in xrange(1,len(rev)+1):
      if rev[i] in w1 and w1[rev[i]] != 0:
         lines.write("%.16f\n"%w1[rev[i]])
      else:
         lines.write("0\n")
   
   for i in xrange(1,len(rev)+1):
      if rev[i] in w2 and w2[rev[i]] != 0:
         lines.write("%.16f\n"%w2[rev[i]])
      else:
         lines.write("0\n")

   lines.close()




from multiprocessing import Pool
import subprocess
from pyprove import expres, eprover
import os

def proofstate(f_dat, f_pos, f_neg, hashing=None):
   def parse(clause):
      clause = clause[clause.rindex("proofvector")+12:].rstrip(",\n").strip().split(",")
      clause = [x.split("(")[0].split(":") for x in clause if x]
      if not hashing:
         clause = ["$%s/%s"%tuple(x) for x in clause if x]
      else:
         clause = ["%s:%s"%tuple(x) for x in clause if x]
      return " ".join(clause)
   dat = file(f_dat).read().strip().split("\n")
   dat = [x for x in dat if x]
   i = 0
   for pos in file(f_pos):
      dat[i] += " "+parse(pos)
      i += 1
   for neg in file(f_neg):
      dat[i] += " "+parse(neg)
      i += 1
   if i != len(dat):
      raise Exception("File %s does not match files %s and %s!" % (f_dat,f_pos,f_neg))
   file(f_dat, "w").write("\n".join(dat))

def prepare1(job):
   (bid,pid,problem,limit,version,force,hashing) = job

   f_problem = expres.benchmarks.path(bid, problem)
   f_cnf = expres.benchmarks.path(bid, "."+problem)+".cnf"
   if not os.path.isfile(f_cnf):
      file(f_cnf, "w").write(eprover.runner.cnf(f_problem))

   result = None
   #result = rkeys[(bid,pid,problem,limit)]
   f_pos = expres.results.path(bid, pid, problem, limit, ext="pos")
   f_neg = expres.results.path(bid, pid, problem, limit, ext="neg")
   if force or (not (os.path.isfile(f_pos) and os.path.isfile(f_neg))):
      result = expres.results.load(bid, pid, problem, limit, trains=True, proof=True)
      if force or not os.path.isfile(f_pos):
         file(f_pos, "w").write("\n".join(result["POS"]))
      if force or not os.path.isfile(f_neg):
         file(f_neg, "w").write("\n".join(result["NEG"]))
      # extract additional positive samples from the proof
      #f_sol = expres.results.path(bid, pid, problem, limit, ext="sol")
      #file(f_sol, "w").write("\n".join(result["PROOF"]))
      #f_prf = expres.results.path(bid, pid, problem, limit, ext="prf")
      #prf = file(f_prf, "w")
      #subprocess.call(["eprover", "--free-numbers", "--cnf", f_sol], stdout=prf)
      ##subprocess.call(["eprover", "--free-numbers", "--cnf", "--no-preprocessing", f_sol], stdout=prf)
      #prf.close()
      #os.system("cat %s | grep '^cnf' >> %s" % (f_prf, f_pos))
   
   f_dat = expres.results.path(bid, pid, problem, limit, ext="in" if hashing else "pre")
   if force or not os.path.isfile(f_dat):
      out = file(f_dat, "w")
      if not hashing:
         subprocess.call(["enigma-features", "--free-numbers", "--enigma-features=%s"%version, \
            f_pos, f_neg, f_cnf], stdout=out)
            #stdout=out, stderr=subprocess.STDOUT)
      else:
         subprocess.call(["enigma-features", "--free-numbers", "--enigma-features=%s"%version, \
            "--feature-hashing=%s"%hashing, f_pos, f_neg, f_cnf], stdout=out)

      out.close()
      if "W" in version:
         proofstate(f_dat, f_pos, f_neg, hashing)

def prepare(rkeys, version, force=False, cores=1, hashing=None):
   jobs = [rkey+(version,force,hashing) for rkey in rkeys]
   pool = Pool(cores)
   res = pool.map_async(prepare1, jobs).get(365*24*3600)
   pool.close()

def translate(f_cnf, f_conj, f_out):
   "deprecated?"

   out = file(f_out, "w")
   if not f_conj:
      subprocess.call(["enigma-features", "--free-numbers", f_cnf], stdout=out)
   else:   
      f_empty = "empty.tmp"
      os.system("rm -fr %s" % f_empty)
      os.system("touch %s" % f_empty)
      subprocess.call(["enigma-features", "--free-numbers", f_cnf, f_empty, f_conj], \
         stdout=out)
         #stdout=out, stderr=subprocess.STDOUT)
      os.system("rm -fr %s" % f_empty)
   out.close()

def make(rkeys, out=None, hashing=None):
   dat = []
   for (bid, pid, problem, limit) in rkeys:
      f_dat = expres.results.path(bid, pid, problem, limit, ext="in" if hashing else "pre")
      if out:
         tmp = file(f_dat).read().strip()
         if tmp:
            out.write(tmp)
            out.write("\n")
      else:
         dat.extend(file(f_dat).read().strip().split("\n"))
   return dat if not out else None




from setuptools import setup, find_packages

setup(name='enigmatic',
      version='0.1',
      description='Enigma: Automatic Inference Guiding Machine',
      url='http://github.com/ai4reason/enigmatic',
      author='ai4reason',
      license='GPL3',
      packages=find_packages(),
      scripts=[
         'bin/eprover',
         'bin/enigma-features',
         'bin/train',
         'bin/predict'
      ],
      install_requires=[
         'xgboost',
         'lightgbm',
         'pyprove==0.1'
      ],
      zip_safe=False)




import math

from .enigmap import fhash

PREFIX = {
   "+": "+1",
   "-": "+0",
   "*": ""
}

BOOSTS = {
   "WRONG:POS": 1,
   "WRONG:NEG": 0
}

def count(ftrs, vector, emap, offset, strict=True):
   for ftr in ftrs:
      if not ftr:
         continue
      if ftr.startswith("$"):
         continue
      if "/" in ftr:
         parts = ftr.split("/")
         ftr = parts[0]
         inc = int(parts[1])
         if inc == 0:
            continue
      else:
         inc = 1
      if (not strict) and (ftr not in emap):
         continue
      if isinstance(emap, int): # hashing version (emap::int is the base)
         fid = fhash(ftr, emap) + offset
      else:
         fid = emap[ftr] + offset
      vector[fid] = vector[fid]+inc if fid in vector else inc

def proofstate(ftrs, vector, offset, hashing=None):
   # TODO: hash also proof numbers (modulo hashing)
   for ftr in ftrs:
      if not ftr.startswith("$"):
         continue
      (num, val) = ftr[1:].split("/")
      (num, val) = (int(num), float(val))
      if not val:
         continue
      vector[offset+num] = val

def string(sign, vector):
   ftrs = ["%s:%s"%(fid,vector[fid]) for fid in sorted(vector)] 
   ftrs = "%s %s"%(PREFIX[sign], " ".join(ftrs))
   return ftrs

def normalize(vector):
    non0 = len([x for x in vector if vector[x]])
    non0 = math.sqrt(non0)
    return {x:vector[x]/non0 for x in vector}

def encode(pr, emap, strict=True):
   (sign,clause,conj) = pr.strip().split("|")
   vector = {}
   count(clause.strip().split(" "), vector, emap, 0, strict)
   conjs = conj.strip().split(" ")
   base = emap if isinstance(emap,int) else len(emap)
   count(conjs, vector, emap, base, strict)
   proofstate(conjs, vector, 2*base, emap)
   #vector = normalize(vector)
   return string(sign, vector)

def make(pre, emap, out=None, strict=True):
   train = []
   for pr in pre:
      tr = encode(pr, emap, strict)
      if out:
         out.write(tr)
         out.write("\n")
      else:
         train.append(tr)
   return train if not out else None

def boost(f_in, f_out, out, method="WRONG:POS"):
   if method not in BOOSTS:
      raise Exception("Unknown boost method (%s)")
   CLS = BOOSTS[method]

   ins = file(f_in).read().strip().split("\n")
   outs = file(f_out).read().strip().split("\n")

   for (correct,predicted) in zip(ins,outs):
      out.write(correct)
      out.write("\n")
      cls = int(correct.split()[0])
      if cls == CLS and cls != int(predicted):
         out.write(correct)
         out.write("\n")




#!/usr/bin/python

import sys

class Learner:
   
   def __init__(self):
      pass

   def efun(self):
      "E Prover weight function name."
      return "Enigma"

   def ext(self):
      "Model filename extension."
      return "ext"

   def name(self):
      return "learner"

   def train(self, f_in, f_mod):
      pass

   def build(self, f_in, f_mod, log=None):
      if log:
         log.write("\nTraining Enigma model (%s):\n\n" % f_in)
         oldout = sys.stdout
         olderr = sys.stderr
         sys.stdout = log
         sys.stderr = log
      ret = self.train(f_in, f_mod)
      if log:
         sys.stdout = oldout
         sys.stderr = olderr
      return ret

   def predict(self, f_in, f_mod):
      return {}

